import torch
from escnn import gspaces
from escnn import nn


def create_block(in_type, out_type, kernel_size, padding, stride):
    return nn.SequentialModule(
        nn.R2Conv(in_type,
                  out_type,
                  kernel_size=kernel_size,
                  padding=padding,
                  stride=stride,
                  bias=False), nn.InnerBatchNorm(out_type),
        nn.ReLU(out_type, inplace=True))


def compute_new_features(nfeatures, kernel_size, padding, stride):
    # **Out dim of a convolution is:
    # (in_image_res - kernel_size + 2*pad)/stride + 1
    return int((nfeatures - kernel_size + 2 * padding) / stride) + 1


def create_pool(out_type, sigma, padding, stride):
    return nn.SequentialModule(
        nn.PointwiseAvgPoolAntialiased(out_type,
                                       sigma=sigma,
                                       stride=stride,
                                       padding=padding))


def create_fully_connected(c, nscreens, final_n_features=64):
    return torch.nn.Sequential(
        torch.nn.Linear(c, final_n_features),
        torch.nn.BatchNorm1d(final_n_features),
        torch.nn.ELU(inplace=True),
        torch.nn.Linear(final_n_features, nscreens),
        torch.nn.Sigmoid(),
    )


class SteerableCNN(torch.nn.Module):

    def __init__(self, config, symmetry):

        super(SteerableCNN, self).__init__()

        self.nscreens = config['speckle']['nscreens']
        self.in_image_res = config['preproc']['resize']
        self.ensemble = config['preproc']['ensemble']

        self.KERNEL_SIZES = config['scnn']['KERNEL_SIZES']
        self.PADDINGS = config['scnn']['PADDINGS']
        self.STRIDES = config['scnn']['STRIDES']
        self.FEATURE_FIELDS = config['scnn']['FEATURE_FIELDS']
        self.POOL_INDICES = config['scnn']['POOL_INDICES']
        self.SIGMA = config['scnn']['SIGMA']
        self.POOL_STRIDES = config['scnn']['POOL_STRIDES']
        self.POOL_PADDINGS = config['scnn']['POOL_PADDINGS']
        self.final_n_features = config['scnn']['final_n_features']

        # Compute the size of the feature map
        self.nfeatures = self.in_image_res
        if symmetry == 'C8':
            # the model is equivariant under rotations by 45 degrees, modelled by C8
            self.r2_act = gspaces.rot2dOnR2(N=8)
        elif symmetry == 'C16':
            # the model is equivariant under rotations by 22.5 degrees, modelled by C16
            self.r2_act = gspaces.rot2dOnR2(N=16)
        elif symmetry == 'C4':
            # the model is equivariant under rotations by 90 degrees, modelled by C4
            self.r2_act = gspaces.rot2dOnR2(N=4)
        else:
            raise ValueError('The symmetry must be C8, C16 or C4')

        # The input image is a scalar field, corresponding to the trivial representation
        in_type = nn.FieldType(self.r2_act, [self.r2_act.trivial_repr])
        # we store the input type for wrapping the images into a geometric tensor during the forward pass
        self.input_type = in_type
        # This mask is to remove the 'corner' pixels in the input image, that would get outside of the square while rotating
        self.mask = nn.MaskModule(in_type, self.in_image_res, margin=0),

        self.blocks = torch.nn.ModuleList()
        for i in range(len(self.KERNEL_SIZES)):
            # Specify the output type of the convolutional layer
            # we choose feature fields, each transforming under the regular representation of the group
            out_type = nn.FieldType(
                self.r2_act,
                self.FEATURE_FIELDS[i] * [self.r2_act.regular_repr])
            block = create_block(in_type, out_type, self.KERNEL_SIZES[i],
                                 self.PADDINGS[i], self.STRIDES[i])
            self.blocks.append(block)
            in_type = block.out_type
            self.nfeatures = compute_new_features(self.nfeatures,
                                                  self.KERNEL_SIZES[i],
                                                  self.PADDINGS[i],
                                                  self.STRIDES[i])
            # Once every two layers, add a pooling layer
            if i in self.POOL_INDICES:
                j = self.POOL_INDICES.index(i)
                pool = create_pool(out_type, self.SIGMA[j],
                                   self.POOL_PADDINGS[j], self.POOL_STRIDES[j])
                self.blocks.append(pool)
                # For gaussian blur filters, the kernel size is 2 * int(round(3 * sigma)) + 1
                kernel_size = 2 * int(round(3 * self.SIGMA[j])) + 1
                self.nfeatures = compute_new_features(self.nfeatures,
                                                      kernel_size,
                                                      self.POOL_PADDINGS[j],
                                                      self.POOL_STRIDES[j])
        if self.nfeatures < 1:
            raise ValueError(
                'The number of features is too small. Please, check the kernel sizes, paddings and strides'
            )

        # Apply a group pooling layer to impose the invariance
        self.gpool = nn.GroupPooling(out_type)

        # number of output channels
        c = self.gpool.out_type.size * self.nfeatures * self.nfeatures

        # If the model uses multiple images as input, I add an extra channel as confidence weight to average the final prediction
        if self.ensemble > 1:
            out_size = self.nscreens + 1
        else:
            out_size = self.nscreens

        self.fully_net = create_fully_connected(c, out_size,
                                                self.final_n_features)

    def forward(self, input: torch.Tensor):
        # wrap the input tensor in a GeometricTensor
        # (associate it with the input type)
        x = nn.GeometricTensor(input, self.input_type)

        # apply each equivariant block
        for block in self.blocks:
            x = block(x)

        # pool over the group
        x = self.gpool(x)

        # unwrap the output GeometricTensor
        # (take the Pytorch tensor and discard the associated representation)
        x = x.tensor

        # classify with the final fully connected layers
        x = self.fully_net(x.reshape(x.shape[0], -1))

        return x



import numpy as np
import torch
from typing import Callable
import matplotlib.pyplot as plt
from speckcn2.utils import ensure_directory


def score_plot(
    conf: dict,
    inputs: torch.Tensor,
    outputs: torch.Tensor,
    tags: list,
    loss: torch.Tensor,
    i: int,
    counter: int,
    recover_tag: list[Callable],
) -> None:
    """Plots side by side the input image, the predicted/exact tags and their
    normalize value in model units.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    inputs : torch.Tensor
        The input speckle patterns
    outputs : torch.Tensor
        The predicted screen tags
    tags : list
        The exact tags of the data
    loss : torch.Tensor
        The loss of the model
    i : int
        The batch index of the image
    counter : int
        The global index of the image
    recover_tag : list
        List of functions to recover each tag
    """
    model_name = conf['model']['name']
    data_dir = conf['speckle']['datadirectory']
    ensemble = conf['preproc']['ensemble']

    # Plot the image and output side by side
    fig, axs = plt.subplots(1, 2 + ensemble, figsize=(4 * (2 + ensemble), 3.5))
    for n in range(ensemble):
        axs[n].imshow(inputs[ensemble * i + n].detach().cpu().squeeze(),
                      cmap='bone')
    axs[0].set_title('Loss: {:.4f}'.format(loss.item()))
    recovered_tag_true = np.asarray([
        recover_tag[j](tags[i][j].detach().cpu().numpy(), i)
        for j in range(len(tags[i]))
    ])
    axs[-2].plot(10**(recovered_tag_true), 'o', label='True')
    recovered_tag_model = np.asarray([
        recover_tag[j](outputs[i][j].detach().cpu().numpy(), i)
        for j in range(len(tags[i]))
    ])
    axs[-2].plot(10**(recovered_tag_model),
                 '.',
                 color='tab:red',
                 label='Predicted')
    axs[-2].set_yscale('log')
    axs[-2].set_title('Screen Tags')
    axs[-2].legend()
    axs[-1].plot(tags[i].detach().cpu().numpy(), 'o', label='True')
    axs[-1].plot(outputs[i].detach().cpu().numpy(),
                 '.',
                 color='tab:red',
                 label='Predicted')
    axs[-1].set_title('Unnormalized out')
    axs[-1].set_ylim(0, 1)
    axs[-1].legend()
    # tight layout
    plt.tight_layout()
    plt.savefig(f'{data_dir}/{model_name}_score/{counter}.png')
    plt.close()


def altitude_profile_plot(
    conf: dict,
    inputs: torch.Tensor,
    outputs: torch.Tensor,
    tags: list,
    i: int,
    counter: int,
    recover_tag: list[Callable],
) -> None:
    """Plots side by side the input image and the predicted/exact altitude
    profile.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    inputs : torch.Tensor
        The input speckle patterns
    outputs : torch.Tensor
        The predicted screen tags
    tags : list
        The exact tags of the data
    i : int
        The batch index of the image
    counter : int
        The global index of the image
    recover_tag : list
        List of functions to recover each tag
    """
    model_name = conf['model']['name']
    data_dir = conf['speckle']['datadirectory']
    ensemble = conf['preproc']['ensemble']

    # Plot the image and output side by side
    fig, axs = plt.subplots(1, 1 + ensemble, figsize=(4 * (1 + ensemble), 3.5))
    for n in range(ensemble):
        axs[n].imshow(inputs[ensemble * i + n].detach().cpu().squeeze(),
                      cmap='bone')
    recovered_tag_true = np.asarray([
        recover_tag[j](tags[i][j].detach().cpu().numpy(), i)
        for j in range(len(tags[i]))
    ])
    h_true = 10**(recovered_tag_true)
    recovered_tag_model = np.asarray([
        recover_tag[j](outputs[i][j].detach().cpu().numpy(), i)
        for j in range(len(tags[i]))
    ])
    h_predic = 10**(recovered_tag_model)

    axs[-1].plot(h_true, range(len(h_true)), 'o--', label='True')

    plot_model_prediciton = False
    if plot_model_prediciton:
        axs[-1].plot(h_predic,
                     range(len(h_predic)),
                     '.',
                     color='tab:red',
                     label='Predicted')

    axs[-1].set_xscale('log')
    axs[-1].set_xlabel(r'$C_{n}^{2}$')
    axs[-1].set_ylabel('Altitude [m]')
    y_tick_labels = [f'{i*400}m' for i in range(0, 8)]
    axs[-1].yaxis.set_ticks(range(0, 8))
    axs[-1].set_yticklabels(y_tick_labels)

    plt.suptitle(r'$C_{n}^{2}(h)$ profile from Speckle Pattern')
    plt.tight_layout()
    plt.savefig(f'{data_dir}/{model_name}_score/h_{counter}.png')
    plt.close()


def plot_loss(conf: dict, model, data_dir: str) -> None:
    """Plots the loss of the model.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    model : torch.nn.Module
        The model to plot the loss of
    data_dir : str
        The directory where the data is stored
    """

    model_name = conf['model']['name']

    ensure_directory(f'{data_dir}/result_plots')

    # plot the loss
    fig, axs = plt.subplots(1, 1, figsize=(5, 5))
    axs.plot(model.epoch, model.loss, label='Training loss')
    axs.plot(model.epoch, model.val_loss, label='Validation loss')
    axs.set_xlabel('Epoch')
    axs.set_ylabel('Loss')
    axs.legend()
    plt.title(f'Model: {model_name}')
    plt.tight_layout()
    plt.savefig(f'{data_dir}/result_plots/loss_{model_name}.png')
    plt.close()


def plot_time(conf: dict, model, data_dir: str) -> None:
    """Plots the time per epoch of the model.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    model : torch.nn.Module
        The model to plot the loss of
    data_dir : str
        The directory where the data is stored
    """

    model_name = conf['model']['name']

    ensure_directory(f'{data_dir}/result_plots')

    # plot the loss
    fig, axs = plt.subplots(1, 1, figsize=(5, 5))
    axs.plot(model.epoch, model.time, label='Time per epoch')
    axs.set_xlabel('Epoch')
    axs.set_ylabel('Time [s]')
    axs.legend()
    plt.title(f'Model: {model_name}')
    plt.tight_layout()
    plt.savefig(f'{data_dir}/result_plots/time_{model_name}.png')
    plt.close()



import torch
import numpy as np
import itertools
import torchvision
from torch import nn
from speckcn2.io import load_model_state
from speckcn2.scnn import SteerableCNN


def setup_model(config: dict) -> tuple[nn.Module, int]:
    """Returns the model specified in the configuration file, with the last
    layer corresponding to the number of screens.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration

    Returns
    -------
    model : torch.nn.Module
        The model with the loaded state
    last_model_state : int
        The number of the last model state
    """

    model_name = config['model']['name']
    model_type = config['model']['type']
    pretrained = config['model']['pretrained']
    nscreens = config['speckle']['nscreens']
    data_directory = config['speckle']['datadirectory']
    img_res = config['preproc']['resize']
    ensemble = config['preproc']['ensemble']

    print(f'^^^ Initializing model {model_name} of type {model_type}')

    if model_type in ['resnet18', 'resnet50', 'resnet152']:
        return get_a_resnet(nscreens, data_directory, model_name, model_type,
                            pretrained, ensemble)
    if model_type in ['scnnC8', 'scnnC16', 'scnnC4']:
        return get_scnn(config)
    else:
        raise ValueError(f'Unknown model {model_name}')


def get_a_resnet(nscreens: int,
                 datadirectory: str,
                 model_name: str,
                 model_type: str,
                 pretrained: bool,
                 ensemble: int = 1) -> tuple[nn.Module, int]:
    """Returns a pretrained ResNet model, with the last layer corresponding to
    the number of screens.

    Parameters
    ----------
    nscreens : int
        Number of screens
    datadirectory : str
        Path to the directory containing the data
    model_name : str
        The name of the model
    model_type : str
        The type of the ResNet
    pretrained : bool
        Whether to use a pretrained model or not
    ensemble : int
        The number of input images that will be processd together as an ensemble corresponding to the same output

    Returns
    -------
    model : torch.nn.Module
        The model with the loaded state
    last_model_state : int
        The number of the last model state
    """

    if model_type == 'resnet18':
        model = torchvision.models.resnet18(
            weights='IMAGENET1K_V1' if pretrained else None)
        finaloutsize = 512
    elif model_type == 'resnet50':
        model = torchvision.models.resnet50(
            weights='IMAGENET1K_V2' if pretrained else None)
        finaloutsize = 2048
    elif model_type == 'resnet152':
        model = torchvision.models.resnet152(
            weights='IMAGENET1K_V2' if pretrained else None)
        finaloutsize = 2048
    else:
        raise ValueError(f'Unknown model {model_type}')

    # If the model uses multiple images as input, I add an extra channel as confidence weight to average the final prediction
    if ensemble > 1:
        nscreens = nscreens + 1

    # Give it its name
    model.name = model_name

    # Change the model to process black and white input
    model.conv1 = torch.nn.Conv2d(1,
                                  64,
                                  kernel_size=(7, 7),
                                  stride=(2, 2),
                                  padding=(3, 3),
                                  bias=False)
    # Add a final fully connected piece to predict the output
    model.fc = torch.nn.Sequential(
        torch.nn.Linear(finaloutsize, 64),
        torch.nn.BatchNorm1d(64),
        torch.nn.ELU(inplace=True),
        torch.nn.Linear(64, nscreens),
        torch.nn.Sigmoid(),
    )

    return load_model_state(model, datadirectory)


class EnsembleModel(nn.Module):
    """Wrapper that allows any model to be used for ensembled data."""

    def __init__(self,
                 ensemble_size: int,
                 device: torch.device,
                 uniform_ensemble: bool = False):
        """Initializes the EnsembleModel.

        Parameters
        ----------
        ensemble_size : int
            The number of images that will be processed together as an ensemble
        device : torch.device
            The device to use
        uniform_ensemble : bool
            If true, all the images in the ensemble will have the same weight
        """
        super(EnsembleModel, self).__init__()
        self.ensemble_size = ensemble_size
        self.device = device
        self.uniform_ensemble = uniform_ensemble

    def forward(self, model, batch_ensemble):
        """Forward pass through the model.

        Parameters
        ----------
        model : torch.nn.Module
            The model to use
        batch_ensemble : list
            Each element is a batch of an ensemble of samples.
        """

        if self.ensemble_size == 1:
            batch = batch_ensemble
            # If no ensembling, each element of the batch is a tuple (image, tag, ensemble_id)
            images, tags, ensembles = zip(*batch)
            images = torch.stack(images).to(self.device)
            tags = torch.tensor(np.stack(tags)).to(self.device)

            return model(images), tags, images
        else:
            batch = list(itertools.chain(*batch_ensemble))
            # Like the ensemble=1 case, I can process independently each element of the batch
            images, tags, ensembles = zip(*batch)
            images = torch.stack(images).to(self.device)
            tags = torch.tensor(np.stack(tags)).to(self.device)

            model_output = model(images)

            # To average the self.ensemble_size outputs of the model I extract the confidence weights
            predictions = model_output[:, :-1]
            weights = model_output[:, -1]
            if self.uniform_ensemble:
                weights = torch.ones_like(weights)
            # multiply the prediction by the weights
            weighted_predictions = predictions * weights.unsqueeze(-1)
            # and sum over the ensembles
            weighted_predictions = weighted_predictions.view(
                model_output.size(0) // self.ensemble_size, self.ensemble_size,
                -1).sum(dim=1)
            # then normalize by the sum of the weights
            sum_weights = weights.view(
                weights.size(0) // self.ensemble_size,
                self.ensemble_size).sum(dim=1)
            ensemble_output = weighted_predictions / sum_weights.unsqueeze(-1)

            # and get the tags and ensemble_id of the first element of the ensemble
            tags = tags[::self.ensemble_size]
            ensembles = ensembles[::self.ensemble_size]

            return ensemble_output, tags, images


def get_scnn(config: dict) -> tuple[nn.Module, int]:
    """Returns a pretrained Spherical-CNN model, with the last layer
    corresponding to the number of screens."""

    model_name = config['model']['name']
    model_type = config['model']['type']
    datadirectory = config['speckle']['datadirectory']

    if model_type == 'scnnC8':
        scnn_model = SteerableCNN(config, 'C8')
    elif model_type == 'scnnC16':
        scnn_model = SteerableCNN(config, 'C16')
    elif model_type == 'scnnC4':
        scnn_model = SteerableCNN(config, 'C4')
    else:
        raise ValueError(f'Unknown model {model_type}')
    scnn_model.name = model_name

    return load_model_state(scnn_model, datadirectory)



import torch
from dataclasses import dataclass
from PIL import Image
import os
import numpy as np
from typing import Callable
import torchvision.transforms as transforms
from speckcn2.utils import ensure_directory, plot_preprocessed_image
from speckcn2.transformations import PolarCoordinateTransform, ShiftRowsTransform, ToUnboundTensor


def assemble_transform(conf: dict) -> transforms.Compose:
    """Assembles the transformation to apply to each image.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration

    Returns
    -------
    transform : torchvision.transforms.Compose
        Transformation to apply to the images
    """
    list_transforms = []

    if conf['preproc']['centercrop']:
        # Take only the center of the image
        list_transforms.append(
            transforms.CenterCrop(conf['preproc']['centercrop']))

    if conf['preproc']['polarize']:
        # make the image larger for better polar conversion
        list_transforms.append(transforms.Resize(conf['preproc']['polresize']))
        # Convert to polar coordinates
        list_transforms.append(PolarCoordinateTransform())

    if conf['preproc']['randomrotate'] and not conf['preproc']['polarize']:
        # Randomly rotate the image, since it is symmetric (not if it is polarized)
        list_transforms.append(transforms.RandomRotation(degrees=(-180, 180)))

    if conf['preproc']['resize']:
        # Optionally, downscale it
        list_transforms.append(
            transforms.Resize(
                (conf['preproc']['resize'], conf['preproc']['resize'])))

    if conf['preproc']['equivariant']:
        # Apply the equivariant transform
        list_transforms.append(ShiftRowsTransform())

    list_transforms.append(ToUnboundTensor())

    return transforms.Compose(list_transforms)


def prepare_data(
    conf: dict,
    nimg_print: int = 5,
) -> tuple[list, list, list]:
    """If not already available, preprocesses the data by loading images and
    tags from the given directory, applying a transformation to the images.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    nimg_print: int
        Number of images to print

    Returns
    -------
    all_images : list
        List of all images
    all_tags : list
        List of all tags
    all_ensemble_ids : list
        List of all ensemble ids, representing images from the same Cn2 profile
    """
    datadirectory = conf['speckle']['datadirectory']
    dataname = conf['preproc']['dataname']
    tagname = dataname.replace('images', 'tags')
    ensemblename = dataname.replace('images', 'ensemble')

    # First, check if the data has already been preprocessed
    if os.path.exists(os.path.join(datadirectory, dataname)):
        print(f'*** Loading preprocessed data from {dataname}')
        # If so, load it
        all_images = torch.load(os.path.join(datadirectory, dataname))
        all_tags = torch.load(os.path.join(datadirectory, tagname))
        all_ensemble_ids = torch.load(os.path.join(datadirectory,
                                                   ensemblename))
    else:
        # Otherwise, preprocess the raw data separating the single images
        all_images, all_tags, all_ensemble_ids = imgs_as_single_datapoint(
            conf, nimg_print)

    return all_images, all_tags, all_ensemble_ids


def imgs_as_single_datapoint(
    conf: dict,
    nimg_print: int = 5,
) -> tuple[list, list, list]:
    """Preprocesses the data by loading images and tags from the given
    directory, applying a transformation to the images. Each image is treated
    as a single data point.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    nimg_print: int
        Number of images to print

    Returns
    -------
    all_images : list
        List of all images
    all_tags : list
        List of all tags
    all_ensemble_ids : list
        List of all ensemble ids, representing images from the same Cn2 profile
    """
    datadirectory = conf['speckle']['datadirectory']
    mname = conf['model']['name']
    dataname = conf['preproc']['dataname']
    tagname = dataname.replace('images', 'tags')
    ensemblename = dataname.replace('images', 'ensemble')
    nreps = conf['preproc']['speckreps']

    # Dummy transformation to get the original image
    transform_orig = transforms.Compose([
        transforms.CenterCrop(conf['preproc']['centercrop']),
        transforms.ToTensor(),
    ])

    # and get the transformation to apply to each image
    transform = assemble_transform(conf)

    # Get the list of images:
    file_list = [
        file_name for file_name in os.listdir(datadirectory)
        if '.txt' in file_name and 'tag' not in file_name
    ]
    # and randomly shuffle them
    np.random.shuffle(file_list)
    # I can also do data augmentation, by using each file multiple times (only if I am also doing a random rotation)
    file_list = file_list * nreps
    all_images = []
    all_tags = []
    all_ensemble_ids = []
    if nimg_print > 0:
        show_image = True
        # and create a folder to store the images
        ensure_directory(f'{datadirectory}/imgs_to_{mname}')
    else:
        show_image = False

    # Check existence of tag files
    tag_files = {}
    for file_name in file_list:
        if 'MALES' in file_name:
            ftagname = file_name.replace('.txt', '_tag.txt')
        else:
            base_name, _, _ = file_name.rpartition('_')
            ftagname = base_name + '_tag.txt'

        tag_path = os.path.join(datadirectory, ftagname)
        if os.path.exists(tag_path):
            tag_files[file_name] = tag_path
        else:
            print(f'*** Warning: tag file {ftagname} not found.')

    # I associate each Cn2 profile to an ensemble ID for parallel processing
    ensembles = {}
    ensemble_counter = 1
    for value in tag_files.values():
        # Check if the value is already assigned an ID
        if value not in ensembles:
            # Assign a new ID if it's a new value
            ensembles[value] = ensemble_counter
            ensemble_counter += 1
    # Create a new dictionary to associate IDs with original values
    ensemble_dict = {key: ensembles[value] for key, value in tag_files.items()}

    # Load each text file as an image
    for counter, file_name in enumerate(file_list):
        # Process only if tags available
        if file_name in tag_files:

            # Construct the full path to the file
            file_path = os.path.join(datadirectory, file_name)

            # Open the text file as an image using PIL
            pixel_values = np.loadtxt(file_path,
                                      delimiter=',',
                                      dtype=np.float32)

            # Create the image
            image_orig = Image.fromarray(pixel_values, mode='F')

            # Apply the transformation
            image = transform(image_orig)
            image_orig = transform_orig(image_orig)
            # and add the img to the collection
            all_images.append(image)

            # Load the tags
            tags = np.loadtxt(tag_files[file_name],
                              delimiter=',',
                              dtype=np.float32)

            # Plot the image using maplotlib
            if counter > nimg_print:
                show_image = False
            if show_image:
                plot_preprocessed_image(image_orig, image, tags, counter,
                                        datadirectory, mname, file_name)

            # Preprocess the tags
            np.log10(tags, out=tags)
            # Add the tag to the colleciton
            all_tags.append(tags)

            # Get the ensemble ID
            ensemble_id = ensemble_dict[file_name]
            # and add it to the collection
            all_ensemble_ids.append(ensemble_id)
        else:
            print(f'*** Warning: tag file {ftagname} not found.')

    # Finally, store them before returning
    torch.save(all_images, os.path.join(datadirectory, dataname))
    torch.save(all_tags, os.path.join(datadirectory, tagname))
    torch.save(all_ensemble_ids, os.path.join(datadirectory, ensemblename))

    print('*** Preprocessing complete.', flush=True)

    return all_images, all_tags, all_ensemble_ids


@dataclass
class Normalizer:
    """Class to handle the normalization of images and tags.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    """
    conf: dict

    def normalize_imgs_and_tags(
        self,
        all_images: list[torch.tensor],
        all_tags: list[np.ndarray],
        all_ensemble_ids: list[int],
    ) -> list[tuple[torch.tensor, np.ndarray, int]]:
        """Normalize both the input images and the tags to be between 0 and 1.

        Parameters
        ----------
        all_images : list
            List of all images
        all_tags : list
            List of all tags
        conf : dict
            Dictionary containing the configuration

        Returns
        -------
        dataset : list
            List of tuples (image, normalized_tag)
        """
        # Get a mask for the NaN values
        self._mask_img = 1 - np.isnan(all_images[0])
        # Then replace all the nan values with 0
        all_images = [torch.nan_to_num(image) for image in all_images]

        # Define the normalization functions for the images
        if not hasattr(self, 'normalize_img'):
            # Find the maximum between the maximum and minimum values of the images
            max_img = max([torch.max(image) for image in all_images])
            print('*** Image max:', max_img)
            min_img = min([torch.min(image) for image in all_images])
            print('*** Image min:', min_img)
            range_img = max_img - min_img

            self.normalize_img, self.recover_img = self._img_normalize_functions(
                min_img, range_img)

        # Normalize the images
        normalized_images = [self.normalize_img(image) for image in all_images]

        # Define the normalization functions for the tags
        if not hasattr(self, 'normalize_tag'):
            self.min_tags = np.min(all_tags, axis=0)
            print('*** Tag min:', self.min_tags)
            self.max_tags = np.max(all_tags, axis=0)
            print('*** Tag max:', self.max_tags)

            # get the std deviation if using Z-score
            if self.conf['preproc']['normalization'] == 'zscore':
                self.mean_tags = np.mean(all_tags, axis=0)
                print('*** Tag mean:', self.mean_tags)
                self.std_tags = np.std(all_tags, axis=0)
                print('*** Tag std:', self.std_tags)
            elif self.conf['preproc']['normalization'] == 'unif':
                raise NotImplementedError(
                    '*** uniform normalization is not fully implemented yet. The sorting messes up the ensemble IDs.'
                )
                array_tags = np.array(all_tags)

                # Find the indices that sort the tags
                _sorting_indices = np.argsort(array_tags, axis=0)
                # And the corresponding indices that unsort them
                self._unsorting_indices = np.argsort(np.argsort(array_tags,
                                                                axis=0),
                                                     axis=0)
                self._sorted_tags = np.stack([
                    array_tags[_sorting_indices[:, i], i]
                    for i in range(array_tags.shape[1])
                ]).T
                self.Ndata = array_tags.shape[0]

            self.normalize_tag, self.recover_tag = self._tag_normalize_functions(
            )

        # And normalize the tags
        normalized_tags = np.array([[
            self.normalize_tag[j](tag, tag_id) for j, tag in enumerate(tags)
        ] for tag_id, tags in enumerate(all_tags)],
                                   dtype=np.float32)

        # I can now create the dataset
        dataset = [(image, tag, ensemble_id) for image, tag, ensemble_id in
                   zip(normalized_images, normalized_tags, all_ensemble_ids)]

        return dataset

    def _img_normalize_functions(
            self, min_img: np.ndarray,
            range_img: np.ndarray) -> tuple[Callable, Callable]:
        """Create the normalization and recovery functions for the images. The
        images are normalized between 0 and 1 using global values.

        Parameters
        ----------
        min_img : np.ndarray
            Minimum value for all the images
        range_img : np.ndarray
            Maximum value for all the images

        Returns
        -------
        normalize_fn : Callable
            Function to normalize an image
        recover_fn : Callable
            Function to recover an image
        """
        normalize_fn = (
            lambda x, min_img=min_img, range_img=range_img: self._mask_img *
            (x - min_img) / range_img)

        recover_fn = (
            lambda y, min_img=min_img, range_img=range_img: self._mask_img *
            (y * range_img + min_img))

        return normalize_fn, recover_fn

    def _tag_normalize_functions(
            self) -> tuple[list[Callable], list[Callable]]:
        """Create the normalization and recovery functions for the tags.

        Returns
        -------
        normalize_functions : list
            List of functions to normalize each tag
        recover_functions : list
            List of functions to recover each tag
        """

        if self.conf['preproc']['normalization'] == 'unif':

            normalize_functions = [
                (lambda x, x_id, i=i: self._unsorting_indices[x_id, i] / self.
                 Ndata) for i in range(self.conf['speckle']['nscreens'])
            ]
            recover_functions = [
                (lambda y, i=i: self._sorted_tags[round(y * self.Ndata), i])
                for i in range(self.conf['speckle']['nscreens'])
            ]
            return normalize_functions, recover_functions
        elif self.conf['preproc']['normalization'] == 'zscore':
            normalize_functions = [
                (lambda x, mean=self.mean_tags[i], std=self.std_tags[i]:
                 (x - mean) / std)
                for i in range(self.conf['speckle']['nscreens'])
            ]
            recover_functions = [
                (lambda y, mean=self.mean_tags[i], std=self.std_tags[i]: y *
                 std + mean) for i in range(self.conf['speckle']['nscreens'])
            ]
            return normalize_functions, recover_functions
        elif self.conf['preproc']['normalization'] == 'log':
            # The log normalization follows this formula:
            #   f(x) = (log(x - min + 1) - log(max - min + 1)) / (log(c - min + 1) - log(max - min + 1))
            # where c=1 sets the range of 0<=f(x)<=c=1
            normalize_functions = [
                (lambda x, x_id, min_t=self.min_tags[i], max_t=self.max_tags[
                    i]: np.log((x - min_t + 1) / (max_t - min_t + 1)) / np.log(
                        (2 - min_t) / (max_t - min_t + 1)))
                for i in range(self.conf['speckle']['nscreens'].min_tags)
            ]
            recover_functions = [
                (lambda y, y_id, min_t=self.min_tags[i], max_t=self.max_tags[
                    i]: np.exp(y) * np.log((2 - min_t) / (max_t - min_t + 1)) *
                 (max_t - min_t + 1) + min_t - 1)
                for i in range(self.conf['speckle']['nscreens'].min_tags)
            ]
            return normalize_functions, recover_functions
        elif self.conf['preproc']['normalization'] == 'lin':
            normalize_functions = [
                (lambda x, x_id, min_t=self.min_tags[i], max_t=self.max_tags[
                    i]: (x - min_t) / (max_t - min_t))
                for i in range(self.conf['speckle']['nscreens'])
            ]
            recover_functions = [
                (lambda y, y_id, min_t=self.min_tags[i], max_t=self.max_tags[
                    i]: y * (max_t - min_t) + min_t)
                for i in range(self.conf['speckle']['nscreens'])
            ]
            return normalize_functions, recover_functions
        else:
            raise ValueError(
                f"*** Error in normalization: normalization {self.conf['preproc']['normalization']} unknown."
            )



import torch
import numpy as np
from PIL import Image


class PolarCoordinateTransform(torch.nn.Module):
    """Transform a Cartesian image to polar coordinates."""

    def __init__(self):
        super(PolarCoordinateTransform, self).__init__()

    def forward(self, img):
        """ forward method of the transform
        Args:
            img (PIL Image or Tensor): Image to be scaled.

        Returns:
            PIL Image or Tensor: Rescaled image.
        """

        img = np.array(img)  # Convert PIL image to NumPy array

        # Assuming img is a grayscale image
        height, width = img.shape
        polar_image = np.zeros_like(img)

        # Center of the image
        center_x, center_y = width // 2, height // 2

        # Maximum possible value of r
        max_r = np.sqrt(center_x**2 + center_y**2)

        # Create a grid of (x, y) coordinates
        y, x = np.ogrid[:height, :width]

        # Shift the grid so that the center of the image is at (0, 0)
        x = x - center_x
        y = y - center_y

        # Convert Cartesian to Polar coordinates
        r = np.sqrt(x**2 + y**2)
        theta = np.arctan2(y, x)

        # Rescale r to [0, height)
        r = np.round(r * (height - 1) / max_r).astype(int)

        # Rescale theta to [0, width)
        theta = np.round((theta + 2 * np.pi) % (2 * np.pi) * (width - 1) /
                         (2 * np.pi)).astype(int)

        # Use a 2D histogram to accumulate all values that map to the same polar coordinate
        histogram, _, _ = np.histogram2d(theta.flatten(),
                                         r.flatten(),
                                         bins=[height, width],
                                         range=[[0, height], [0, width]],
                                         weights=img.flatten())

        # Count how many Cartesian coordinates map to each polar coordinate
        counts, _, _ = np.histogram2d(theta.flatten(),
                                      r.flatten(),
                                      bins=[height, width],
                                      range=[[0, height], [0, width]])

        # Take the average of all values that map to the same polar coordinate
        polar_image = histogram / counts

        # Handle any divisions by zero
        polar_image[np.isnan(polar_image)] = 0

        # Crop the large r part that is not used
        for x in range(width - 1, -1, -1):
            # If the column contains at least one non-black pixel
            if np.any(polar_image[:, x] != 0):
                # Crop at this x position
                polar_image = polar_image[:, :x]
                break

        # reconvert to PIL image before returning
        return Image.fromarray(polar_image)


class ShiftRowsTransform(torch.nn.Module):
    """Shift the rows of an image such that the row with the smallest sum is at
    the bottom."""

    def __init__(self):
        super(ShiftRowsTransform, self).__init__()

    def forward(self, img):
        img = np.array(img)  # Convert PIL image to NumPy array

        # Find the row with the largest sum
        row_sums = np.sum(img, axis=1)
        max_sum_row_index = np.argmin(row_sums)

        # Shift all rows respecting periodicity
        shifted_img = np.roll(img, -max_sum_row_index, axis=0)

        # reconvert to PIL image before returning
        return Image.fromarray(shifted_img)


class ToUnboundTensor(torch.nn.Module):
    """Transform the image into a tensor, but do not normalize it like
    torchvision.ToTensor."""

    def __init__(self):
        super(ToUnboundTensor, self).__init__()

    def forward(self, img):
        img = np.array(img)
        # add a color chanel in dim 0
        return torch.from_numpy(img).unsqueeze(0)



import time
import random
import torch
from torch import nn, optim, Tensor
from speckcn2.io import save
from speckcn2.mlmodels import EnsembleModel
from speckcn2.preprocess import Normalizer
from speckcn2.utils import ensure_directory
from speckcn2.plots import score_plot, altitude_profile_plot


def train(model: nn.Module, last_model_state: int, conf: dict, train_set: list,
          test_set: list, device: torch.device, optimizer: optim.Optimizer,
          criterion: nn.Module) -> tuple[nn.Module, float]:
    """Trains the model for the given number of epochs.

    Parameters
    ----------
    model : torch.nn.Module
        The model to train
    last_model_state : int
        The number of the last model state
    conf : dict
        Dictionary containing the configuration
    train_set : list
        The training set
    test_set : list
        The testing set
    device : torch.device
        The device to use
    optimizer : torch.optim
        The optimizer to use
    criterion : torch.nn
        The loss function to use

    Returns
    -------
    model : torch.nn.Module
        The trained model
    average_loss : float
        The average loss of the last epoch
    """

    final_epoch = conf['hyppar']['maxepochs']
    save_every = conf['model']['save_every']
    datadirectory = conf['speckle']['datadirectory']
    batch_size = conf['hyppar']['batch_size']

    # Setup the EnsembleModel wrapper
    ensemble = EnsembleModel(conf['preproc']['ensemble'], device,
                             conf['preproc']['ensemble_unif'])

    print(f'Training the model from epoch {last_model_state} to {final_epoch}')
    average_loss = 0.0
    model.train()
    for epoch in range(last_model_state, final_epoch):
        total_loss = 0.0
        t_in = time.time()
        for i in range(0, len(train_set), batch_size):
            batch = train_set[i:i + batch_size]

            # Zero out the optimizer
            optimizer.zero_grad()

            # Forward pass
            outputs, targets, _ = ensemble(model, batch)
            loss = criterion(outputs, targets)

            # Backward pass
            loss.backward()
            optimizer.step()

            # Accumulate the loss
            total_loss += loss.item()

        # Suffle the training set
        random.shuffle(train_set)

        # Calculate average loss for the epoch
        average_loss = total_loss / len(train_set)

        # Log the important information
        t_fin = time.time() - t_in
        model.loss.append(average_loss)
        model.time.append(t_fin)
        model.epoch.append(epoch + 1)
        # And also the validation loss
        val_loss = 0.0
        with torch.no_grad():
            for i in range(0, len(test_set), batch_size):
                batch = test_set[i:i + batch_size]
                # Forward pass
                outputs, targets, _ = ensemble(model, batch)
                loss = criterion(outputs, targets)
                # sum loss
                val_loss += loss.item()
        val_loss = val_loss / len(test_set)
        model.val_loss.append(val_loss)

        # Print the average loss for every epoch
        print(
            f'Epoch {epoch+1}/{final_epoch} (in {t_fin:.3g}s),\tTrain-Loss: {average_loss:.5f},\tTest-Loss: {val_loss:.5f}',
            flush=True)

        if (epoch + 1) % save_every == 0 or epoch == final_epoch - 1:
            # Save the model state
            save(model, datadirectory)

    return model, average_loss


def score(model: nn.Module,
          test_set: list,
          device: torch.device,
          criterion: nn.Module,
          normalizer: Normalizer,
          nimg_plot: int = 100) -> list[Tensor]:
    """Tests the model.

    Parameters
    ----------
    model : torch.nn.Module
        The model to test
    test_set : list
        The testing set
    device : torch.device
        The device to use
    criterion : torch.nn
        The loss function to use
    normalizer : Normalizer
        The normalizer used to recover the tags
    nimg_plot : int
        Number of images to plot

    Returns
    -------
    test_tags : list
        List of all the predicted tags of the test set
    """
    counter = 0
    conf = normalizer.conf
    data_dir = conf['speckle']['datadirectory']
    batch_size = conf['hyppar']['batch_size']
    # Setup the EnsembleModel wrapper
    ensemble = EnsembleModel(conf['preproc']['ensemble'], device,
                             conf['preproc']['ensemble_unif'])

    with torch.no_grad():
        # Put model in evaluation mode
        model.eval()

        test_tags = []
        # create the directory where the images will be stored
        ensure_directory(f'{data_dir}/{model.name}_score')

        for idx in range(0, len(test_set), batch_size):
            batch = test_set[idx:idx + batch_size]

            # Forward pass
            outputs, targets, inputs = ensemble(model, batch)

            # Loop each input separately
            for i in range(len(outputs)):
                loss = criterion(outputs[i], targets[i])
                # Print the loss for every epoch
                print(f'Item {counter} loss: {loss.item():.4f}')

                if counter < nimg_plot:
                    score_plot(conf, inputs, outputs, targets, loss, i,
                               counter, normalizer.recover_tag)
                    altitude_profile_plot(conf, inputs, outputs, targets, i,
                                          counter, normalizer.recover_tag)

                # and get all the tags for statistic analysis
                for tag in outputs:
                    test_tags.append(tag)

                counter += 1

    return test_tags



import torch
import pickle
import random
import os
import torch.nn as nn
import matplotlib.pyplot as plt


def plot_preprocessed_image(image_orig: torch.tensor, image: torch.tensor,
                            tags: torch.tensor, counter: int,
                            datadirectory: str, mname: str,
                            file_name: str) -> None:
    """Plots the original and preprocessed image, and the tags.

    Parameters
    ----------
    image_orig : torch.tensor
        The original image
    image : torch.tensor
        The preprocessed image
    tags : torch.tensor
        The screen tags
    counter : int
        The counter of the image
    datadirectory : str
        The directory containing the data
    mname : str
        The name of the model
    file_name : str
        The name of the original image
    """

    fig, axs = plt.subplots(1, 3, figsize=(15, 5))
    # Plot the original image
    axs[0].imshow(image_orig.squeeze(), cmap='bone')
    axs[0].set_title(f'Training Image {file_name}')
    # Plot the preprocessd image
    axs[1].imshow(image.squeeze(), cmap='bone')
    axs[1].set_title('Processed as')
    axs[1].set_xlabel(r'$r$')
    axs[1].set_ylabel(r'$\theta$')

    # Plot the tags
    axs[2].plot(tags, 'o')
    axs[2].set_yscale('log')
    axs[2].set_title('Screen Tags')
    axs[2].legend()

    fig.subplots_adjust(wspace=0.3)
    plt.savefig(f'{datadirectory}/imgs_to_{mname}/{counter}.png')
    plt.close()


def ensure_directory(data_directory: str) -> None:
    """Ensure that the directory exists.

    Parameters
    ----------
    data_directory : str
        The directory to ensure
    """

    if not os.path.isdir(data_directory):
        os.mkdir(data_directory)


def setup_loss(config: dict) -> nn.Module:
    """Returns the criterion specified in the configuration file.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration

    Returns
    -------
    criterion : torch.nn.Module
        The criterion with the loaded state
    """

    criterion_name = config['hyppar']['loss']
    if criterion_name == 'BCELoss':
        return torch.nn.BCELoss()
    elif criterion_name == 'MSELoss':
        return torch.nn.MSELoss()
    elif criterion_name == 'MAELoss':
        return torch.nn.L1Loss()
    elif criterion_name == 'Pearson':
        return PearsonCorrelationLoss()
    else:
        raise ValueError(f'Unknown criterion {criterion_name}')


def setup_optimizer(config: dict, model: nn.Module) -> nn.Module:
    """Returns the optimizer specified in the configuration file.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration
    model : torch.nn.Module
        The model to optimize

    Returns
    -------
    optimizer : torch.nn.Module
        The optimizer with the loaded state
    """

    optimizer_name = config['hyppar']['optimizer']
    if optimizer_name == 'Adam':
        return torch.optim.Adam(model.parameters(), lr=config['hyppar']['lr'])
    elif optimizer_name == 'SGD':
        return torch.optim.SGD(model.parameters(), lr=config['hyppar']['lr'])
    else:
        raise ValueError(f'Unknown optimizer {optimizer_name}')


class PearsonCorrelationLoss(nn.Module):

    def __init__(self):
        super(PearsonCorrelationLoss, self).__init__()

    def forward(self, x, y):
        # Calculate mean of x and y
        mean_x = torch.mean(x)
        mean_y = torch.mean(y)

        # Calculate covariance
        cov_xy = torch.mean((x - mean_x) * (y - mean_y))

        # Calculate standard deviations
        std_x = torch.std(x)
        std_y = torch.std(y)

        # Calculate Pearson correlation coefficient
        corr = cov_xy / (std_x * std_y + 1e-8
                         )  # Add a small epsilon to avoid division by zero

        # The loss is 1 - correlation to be minimized
        loss = 1.0 - corr

        return loss


def train_test_split(
    config: dict,
    dataset: list[tuple[torch.tensor, float]],
) -> tuple[list, list]:
    """Splits the data into training and testing sets.

    Parameters
    ----------
    config : dict
        The configuration file
    dataset : list
        List of tuples (image, tag, ensemble_id)

    Returns
    -------
    train_set : list
        Training dataset
    test_set : list
        Testing dataset
    """
    modelname = config['model']['name']
    datadirectory = config['speckle']['datadirectory']
    train_test_split = config['hyppar']['ttsplit']
    ensemble_size = config['preproc']['ensemble']

    # Check if the training and test set are already prepared
    if os.path.isfile(f'{datadirectory}/train_set.pickle') and os.path.isfile(
            f'{datadirectory}/test_set.pickle'):
        train_set = pickle.load(
            open(f'{datadirectory}/train_set_{modelname}.pickle', 'rb'))
        test_set = pickle.load(
            open(f'{datadirectory}/test_set_{modelname}.pickle', 'rb'))
        return train_set, test_set

    # If I am using ensembles, each data point is a tuple of images
    if ensemble_size > 1:
        ensemble_dataset = []

        split_ensembles = {}  # type: dict
        for item in dataset:
            key = item[-1]
            if key not in split_ensembles:
                split_ensembles[key] = []
            split_ensembles[key].append(item)

        for ensemble in split_ensembles:
            # * In each ensemble, take n_groups groups of ensemble_size datapoints
            this_e_size = len(split_ensembles[ensemble])
            n_groups = this_e_size // ensemble_size
            if n_groups < 1:
                raise ValueError(
                    f'Ensemble size {ensemble_size} is too large for ensemble {ensemble} with size {this_e_size}'
                )
            # Extact the ensembles randomly
            sample = random.sample(split_ensembles[ensemble],
                                   n_groups * ensemble_size)
            # split the sample into groups of ensemble_size
            sample = [
                sample[i:i + ensemble_size]
                for i in range(0, n_groups * ensemble_size, ensemble_size)
            ]
            # and append it to the ensemble_dataset as separate elements
            ensemble_dataset.extend(sample)

        # shuffle dimension 0 of the ensemble_dataset
        random.shuffle(ensemble_dataset)

        train_size = int(train_test_split * len(ensemble_dataset))
        train_set = ensemble_dataset[:train_size]
        test_set = ensemble_dataset[train_size:]

        print(
            f'*** There are {len(ensemble_dataset)} ensemble groups in the dataset, that I split in {len(train_set)} for training and {len(test_set)} for testing. Each ensemble is composed by {ensemble_size} images. This corresponds to {len(train_set)*ensemble_size} for training and {len(test_set)*ensemble_size} for testing.'
        )
    else:
        train_size = int(train_test_split * len(dataset))
        print(
            f'*** There are {len(dataset)} images in the dataset, {train_size} for training and {len(dataset)-train_size} for testing.'
        )

        random.shuffle(dataset)
        train_set = dataset[:train_size]
        test_set = dataset[train_size:]

    # Save the training and testing set
    pickle.dump(train_set,
                open(f'{datadirectory}/train_set_{modelname}.pickle', 'wb'))
    pickle.dump(test_set,
                open(f'{datadirectory}/test_set_{modelname}.pickle', 'wb'))

    return train_set, test_set



from setuptools import setup, find_packages

DESCRIPTION = 'Estimate Cn2 from Speckle patterns'
LONG_DESCRIPTION = 'A Python package that uses machine learning to predict turbolence profiles from static patterns'

# Setting up
setup(
    name='speckcn2',
    version='0.0.24',
    author='Simone Ciarella',
    author_email='<s.ciarella@esciencecenter.com>',
    url='https://github.com/MALES-project/SpeckleCn2Profiler',
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    python_requires='>=3.9.0',
    packages=find_packages(),
    install_requires=[
        'numpy', 'matplotlib', 'torch', 'torchvision', 'scipy>=1.10.0',
        'PyYAML', 'escnn'
    ],
    keywords=[
        'python', 'optical satelites', 'machine learning', 'turbolence',
        'laser communication'
    ],
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Science/Research',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Operating System :: Unix',
        'Topic :: Scientific/Engineering',
    ],
)



# This is the __init__.py file for the rkm package

# Import any modules or subpackages here
from .io import save, load, load_config
from .preprocess import prepare_data, Normalizer
from .mlmodels import setup_model, get_a_resnet, EnsembleModel
from .mlops import train, score
from .postprocess import tags_distribution
from .plots import plot_loss, plot_time
from .utils import setup_loss, setup_optimizer, train_test_split

__all__ = [
    'load_config', 'prepare_data', 'Normalizer', 'train_test_split',
    'setup_model', 'get_a_resnet', 'EnsembleModel', 'setup_loss',
    'setup_optimizer', 'train', 'score', 'tags_distribution', 'plot_loss',
    'plot_time', 'save', 'load'
]



import os
import yaml
import torch
from speckcn2.utils import ensure_directory


def load_config(config_file_path: str) -> dict:
    """Load the configuration file.

    Parameters
    ----------
    config_file_path : str
        Path to the .yaml configuration file

    Returns
    -------
    config : dict
        Dictionary containing the configuration
    """
    with open(config_file_path, 'r') as file:
        config = yaml.safe_load(file)
    return config


def save(model: torch.nn.Module, datadirectory: str) -> None:
    """Save the model state and the model itself.

    Parameters
    ----------
    model : torch.nn.Module
        The model to save
    epoch : int
        The epoch of the model
    datadirectory : str
        The directory where the data is stored
    """

    model_state = {
        'epoch': model.epoch,
        'loss': model.loss,
        'val_loss': model.val_loss,
        'time': model.time,
        'model_state_dict': model.state_dict(),
    }

    torch.save(
        model_state,
        f'{datadirectory}/{model.name}_states/{model.name}_{model.epoch[-1]}.pth'
    )


def load(model: torch.nn.Module, datadirectory: str, epoch: int) -> None:
    """Load the model state and the model itself.

    Parameters
    ----------
    model : torch.nn.Module
        The model to load
    datadirectory : str
        The directory where the data is stored
    epoch : int
        The epoch of the model
    """

    model_state = torch.load(
        f'{datadirectory}/{model.name}_states/{model.name}_{epoch}.pth')

    model.epoch = model_state['epoch']
    model.loss = model_state['loss']
    model.val_loss = model_state['val_loss']
    model.time = model_state['time']
    model.load_state_dict(model_state['model_state_dict'], strict=False)

    assert model.epoch[
        -1] == epoch, 'The epoch of the model is not the same as the one loaded'


def load_model_state(model: torch.nn.Module,
                     datadirectory: str) -> tuple[torch.nn.Module, int]:
    """Loads the model state from the given directory.

    Parameters
    ----------
    model : torch.nn.Module
        The model to load the state into
    datadirectory : str
        The directory where the model states are stored

    Returns
    -------
    model : torch.nn.Module
        The model with the loaded state
    last_model_state : int
        The number of the last model state
    """

    # Print model informations
    print(model)
    model.nparams = sum(p.numel() for p in model.parameters())
    print(f'\n--> Nparams = {model.nparams}')

    ensure_directory(f'{datadirectory}/{model.name}_states')

    # check what is the last model state
    try:
        last_model_state = sorted([
            int(file_name.split('.pth')[0].split('_')[-1])
            for file_name in os.listdir(f'{datadirectory}/{model.name}_states')
        ])[-1]
    except Exception as e:
        print(f'Warning: {e}')
        last_model_state = 0

    if last_model_state > 0:
        print(f'Loading model at epoch {last_model_state}')
        load(model, datadirectory, last_model_state)
        return model, last_model_state
    else:
        print('No pretrained model to load')

        # Initialize some model state measure
        model.loss = []
        model.val_loss = []
        model.time = []
        model.epoch = []

        return model, 0



import matplotlib.pyplot as plt
import itertools
import numpy as np
from typing import Callable, Optional
from torch import Tensor
from torch import device as Device
from speckcn2.utils import ensure_directory


def tags_distribution(
        conf: dict,
        train_set: list,
        test_tags: Tensor,
        device: Device,
        rescale: bool = False,
        recover_tag: Optional[list[Callable[[Tensor],
                                            Tensor]]] = None) -> None:
    """Plots the distribution of the tags.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    train_set : list
        The training set
    test_tags : torch.Tensor
        The predicted tags for the test dataset
    device : torch.device
        The device to use
    data_directory : str
        The directory where the data is stored
    rescale : bool, optional
        Whether to rescale the tags using recover_tag() or leave them between 0 and 1
    recover_tag : list, optional
        List of functions to recover each tag
    """

    data_directory = conf['speckle']['datadirectory']
    model_name = conf['model']['name']
    ensemble = conf['preproc']['ensemble']

    ensure_directory(f'{data_directory}/result_plots')

    # Get the tags from the training set
    if ensemble > 1:
        train_set = list(itertools.chain(*train_set))
    _, tags, _ = zip(*train_set)
    tags = np.stack(tags)
    train_tags = np.array([n for n in tags])

    # Get the tags from the test set
    predic_tags = np.array([n.cpu().numpy() for n in test_tags])
    print(f'Data shape: {train_tags.shape}')
    print(f'Prediction shape: {predic_tags.shape}')
    print(f'Train mean: {train_tags.mean()}')
    print(f'Train std: {train_tags.std()}')
    print(f'Prediction mean: {predic_tags.mean()}')
    print(f'Prediction std: {predic_tags.std()}')

    # Plot the distribution of each tag element
    fig, axs = plt.subplots(2, 4, figsize=(20, 10))
    for i in range(train_tags.shape[1]):
        if rescale and recover_tag is not None:
            recovered_tag_model = np.asarray(
                [recover_tag[i](predic_tags[:, i], i)]).squeeze(0)
            print(predic_tags[:, i].shape)
            print(recovered_tag_model.shape)
            recovered_tag_true = np.asarray(
                [recover_tag[i](train_tags[:, i], i)]).squeeze(0)
            axs[i // 4, i % 4].hist(recovered_tag_model,
                                    bins=20,
                                    color='tab:red',
                                    density=True,
                                    alpha=0.5,
                                    label='Model prediction')
            axs[i // 4, i % 4].hist(recovered_tag_true,
                                    bins=20,
                                    color='tab:blue',
                                    density=True,
                                    alpha=0.5,
                                    label='Training data')
        else:
            axs[i // 4, i % 4].hist(predic_tags[:, i],
                                    bins=20,
                                    color='tab:red',
                                    density=True,
                                    alpha=0.5,
                                    label='Model prediction')
            axs[i // 4, i % 4].hist(train_tags[:, i],
                                    bins=20,
                                    color='tab:blue',
                                    density=True,
                                    alpha=0.5,
                                    label='Training data')
        axs[i // 4, i % 4].set_title(f'Tag {i}')
    axs[0, 1].legend()
    plt.tight_layout()
    if rescale:
        plt.savefig(f'{data_directory}/result_plots/{model_name}_tags.png')
    else:
        plt.savefig(
            f'{data_directory}/result_plots/{model_name}_tags_unscaled.png')
    plt.close()
